/*****************************************************
This program was produced by the
CodeWizardAVR V2.03.8a Standard
Automatic Program Generator
 Copyright 1998-2008 Pavel Haiduc, HP InfoTech s.r.l.
http://www.hpinfotech.com

Project : RDS Decoder for ATMega88
Version : 2.00
Date    : 02-06-2012
Author  : Frank Brennecke
Company : 
Comments:


Chip type           : ATmega88
Program type        : Application
Clock frequency     : 20,000000 MHz
Memory model        : Small
External RAM size   : 0
Data Stack size     : 1024
*****************************************************/



#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <string.h>


#define T1CNT           TCNT1H*256 + TCNT1L

#include <lcd-routines.c>

// RDS related procedures

#define INFO "RDS Decoder V2"


// general word on ISR: it does not seem that ISR has global interrupts disabled.

#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <math.h>

#define RDS_PORT PIND
#define RDS_PIN  PD3

volatile unsigned char event = 0;

#define eventTimerOverflow      _BV(0)
// #define eventUsart0RxDInterrupt _BV(1)
#define eventGroupComplete      _BV(2)
#define eventBitstreamSnapshot  _BV(3)
#define eventSyncLost           _BV(4)

volatile unsigned char timerOverflowCount = 0;
volatile unsigned char usart0RxDByte = 0;
volatile unsigned char bitCounter = 0;
volatile unsigned char bitCounterSnapshot = 0;

volatile unsigned char synchronized = 0;
volatile unsigned char syncQuality = 0; //0-31, 0=no useable signal, 15=medium quality, 31=perfect!
volatile unsigned char syncCounter = 0;

volatile unsigned char block = 0;
volatile unsigned long bitstream = 0;
volatile unsigned long bitstreamData = 0;
volatile unsigned int block1 = 0;
volatile unsigned int block2 = 0;
volatile unsigned int block3 = 0;
volatile unsigned int block4 = 0;
volatile unsigned int PositionCounter = 0;
volatile unsigned int radioTextPosition = 0;




// Information Data Set


char RDSLine0[64];
char RDSLine1[64];


//ISR is executed when there is a new rds bit to read
ISR(INT0_vect) 
{
	cli();	
   	unsigned int syndrome = 0;
  
  //Copy global vars to local vars for performance
	unsigned long bitstreamLocal = bitstream;
	unsigned char blockLocal = block;
  
  //shift rds bit in on right side
	bitstreamLocal *= 2; 
	if (RDS_PORT & _BV(RDS_PIN))
		bitstreamLocal++;
	
	if (RDS_PORT & _BV(RDS_PIN))
		PORTD &= ~_BV(PD7);
    else
	    PORTD |= _BV(PD7);  
		
    
	bitCounter++;
  
  //collect data for raw bitstream snapshots
	bitCounterSnapshot++;
	
	if (bitCounterSnapshot == 26) 
	{
		bitstreamData = (bitstreamLocal & 0x3FFFFFF);
		bitCounterSnapshot = 0;
		event |= eventBitstreamSnapshot;
	}
  
  //when we have 26 bits or are not synchronized to the stream
  //check the CRC and maybe store a rds block
	if (bitCounter == 26 || !synchronized) 
	{
  
		bitstreamLocal &= 0x3FFFFFF; //we only need 26 bits
		syndrome = (bitstreamLocal / 0x10000); //bits 16-31 (2^16)
		if (bitstreamLocal & _BV(0))
			syndrome ^= 0x031b;
    
		if (bitstreamLocal & _BV(1))
			syndrome ^= 0x038f;
    
		if (bitstreamLocal & _BV(2))
			syndrome ^= 0x02a7;
      
		if (bitstreamLocal & _BV(3))
			syndrome ^= 0x00f7;
    
		if (bitstreamLocal & _BV(4))
			syndrome ^= 0x01ee;
    
		if (bitstreamLocal & _BV(5))
			syndrome ^= 0x03dc;
		
		if (bitstreamLocal & _BV(6))
			syndrome ^= 0x0201;
		
		if (bitstreamLocal & _BV(7))
			syndrome ^= 0x01bb;
		
		if (bitstreamLocal & _BV(8))
			syndrome ^= 0x0376;
		
		if (bitstreamLocal & _BV(9))
			syndrome ^= 0x0355;
	  
		if (bitstreamLocal & _BV(10))
			syndrome ^= 0x0313;
		
		if (bitstreamLocal & _BV(11))
			syndrome ^= 0x039f;
	  
		if (bitstreamLocal & _BV(12))
			syndrome ^= 0x0287;
		  
		if (bitstreamLocal & _BV(13))
			syndrome ^= 0x00b7;
		
		if (bitstreamLocal & _BV(14))
			syndrome ^= 0x016e;

		if (bitstreamLocal & 0b1000000000000000) // _BV(15) does not work!!!
			syndrome ^= 0x02dc;

    
		if (blockLocal == 0 && syndrome == 0x03d8) 
		{
			//Block A?
			block1 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} 
		else if (blockLocal == 1 && syndrome == 0x03d4) 
		{
			//Block B?
			block2 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} 
		else if (blockLocal == 2 && !(block2 & _BV(11)) && syndrome == 0x025c) 
		{ 
			//Block C type A?
			block3 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} 
		else if (blockLocal == 2 && (block2 & _BV(11)) && syndrome == 0x03cc) 
		{
			//Block C type B?
			block3 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} 
		else if (blockLocal == 3 && syndrome == 0x0258) 
		{
			//Block D?
			block4 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal = 0;
			//we have a complete group!
			event |= eventGroupComplete;
		} 
		else 
		{
      //sync lost..
			synchronized = 0;
			blockLocal = 0;
			event |= eventSyncLost;
		}
		bitCounter = 0; 
	}
  
	if (event & eventGroupComplete) 
	{
//		PORTC |= _BV(PC0);
		if (syncQuality < 31)
			syncQuality++;
    
	} 
	else if (!synchronized) 
	{
//		PORTC &= ~_BV(PC0);
    	syncCounter++;
		if (syncQuality > 0 && syncCounter == 26) 
		{
			syncQuality--;
			syncCounter = 0;
		}
      
	}
	
	if (syncQuality >= 15)
		PORTD &= ~_BV(PD6);
	else
		PORTD |= _BV(PD6);
    
   //Store local vars into global vars to remember state
	bitstream = bitstreamLocal;
	block = blockLocal;
    sei();
}



//timer0 overflowed
ISR(TIMER0_OVF_vect) 
{
	cli();
	timerOverflowCount++;
	if (timerOverflowCount > 4) 
	{
		event |= eventTimerOverflow;
		timerOverflowCount = 0;
	}
	PositionCounter++;
	if (PositionCounter == 48)
	{
		PositionCounter = 0;
		radioTextPosition++;
		if (radioTextPosition == 64)
			radioTextPosition = 0;	
	}		
	sei();
}



void AVR_Init(void)
{
	lcd_init();
	lcd_string(INFO);
    // Port D initialization
    // PD0   out LED4 gelb
    // PD1   out LED5 gelb	
    // PD2   in   INT0/SA user	0x04	
    // PD3   in   INT1/SB user	0x08
    // PD4   out LED2 grn
    // PD5   out LED6 gelb
    // PD6   out LED7 rot
    // PD7   out LED8 rot
	DDRD=0xF3;
    PORTD=0xF3;
	// PC0   out LED1 grn
    // PC1   out LED3 grn	
	DDRC|=0x03;  // LED C0 und C1 auf Ausgang
}


// Main procedure

int main(void) 
{
	
	AVR_Init();

	unsigned int  programmeIdentificationCode = 0;
	unsigned char groupType = 0;
	unsigned char groupVersion = 0;
	unsigned char trafficProgrammeIdentificationCode = 0;
	unsigned char programmeTypeCode = 0;
	unsigned char trafficAnnouncementCode = 0;
	unsigned char musicSpeechSwitchScode = 0;
	unsigned char group0CodeBits = 0;
	unsigned char group0CodeBitsSteadyCount = 0;
	unsigned char programmeServiceName[8];
	unsigned char programmeServiceNameNew[8];
	unsigned char decoderIdentificationControlCode = 0;
	unsigned char decoderIdentificationControlCodeNew = 0;
	unsigned char alternativeFrequencyCodes[27];
	unsigned char syncMessage = 0;
	unsigned char linkageActuator = 0;
	unsigned char extendedCountryCode = 0;
	unsigned char textSegmentAddress = 0;
	unsigned char textSegmentAddressPrevious = 0;
	unsigned char textVersion = 0;
	unsigned char textVersionPrevious = 0; 
	unsigned char radioText[64];
	unsigned char radioTextPrevious[64];
	unsigned char radioTime[32];
	unsigned char radioFrequency[32];
	unsigned char textSegmentAddress0Seen = 0;
	unsigned int  modifiedJulianDay = 0;
	unsigned int  utcYear = 0;
	unsigned char utcMonth = 0;
	unsigned char utcDay = 0;
	signed   char localHours = 0;
	unsigned char utcHours = 0;
	signed   char localMinutes = 0;
	unsigned char utcMinutes = 0;
	unsigned char utcMinutesPrevious = 0xFF;
	unsigned char localSign = 0;
	unsigned int  localTimeOffset = 0;
	unsigned int  radioTextAvailable = 0;
	
	unsigned int  radioFrequencyAvailable = 0;
  
  //general purpose vars
	unsigned int  m;
	unsigned char h;
	unsigned char i;
	unsigned char j;
  


//enable int0
	EIMSK = _BV(INT0);
	EIFR = _BV(INTF0);
  
// INT0 raising edge
	EICRA |= _BV(ISC01) | _BV(ISC00);
  
//timer0 prescaler clk/1024
	TCCR0B |= _BV(CS02) | _BV(CS00);
//enable overflow interrupt
	TIMSK0 |= _BV(TOIE0);
  
	sei(); //enable interrupts;
 
	for (;;) 
	{
		do {} while (!event);
		cli();
		if (event & eventTimerOverflow) 
		{
			event &= ~eventTimerOverflow;
			sei();
		  // no rds signal message
			if (syncQuality == 0 && syncMessage < 12) 
			{
				syncMessage++;
				if (syncMessage == 12) 
				{
					programmeIdentificationCode = 0;
					lcd_clear();
				}	
			}
			  
			if (syncQuality > 0)
				syncMessage = 0;
			  
//			if (PORTC & _BV(PC0))
//				PORTC &= ~_BV(PC0); //heartbeat led off
//			else
//				PORTC |= _BV(PC0); //heartbeat led on
		  
		} 
		else 
		{
		  sei();
		}
		
		//reset vars when sync lost. Warning, sync lost happens 26 times as many as groupcomlete, do not do anything lenghty here
		cli();
		
		if (event & eventSyncLost) 
		{
			event &= ~eventSyncLost;
			sei();
			group0CodeBitsSteadyCount = 5;
		} 
		else 
		{
			sei();
		}
		
		cli();
		if (event & eventGroupComplete) 
		{
			event &= ~eventGroupComplete;
			sei();
			
		  //Group type code
			groupType = block2 / 0x1000; //bits 12-15 (2^12)
		  
		  //Group version code
			if (block2 & _BV(11))
				groupVersion = 'B';
			else
				groupVersion = 'A';
		  
		  //PI Codes of block 3 B-version groups are not decoded.
		  
		  //Programme Identification code
			if (programmeIdentificationCode != block1) 
			{
				lcd_clear();
				programmeIdentificationCode = block1;
			
			/* reset variables because PI code changed */
				trafficProgrammeIdentificationCode = 0xFF;
				programmeTypeCode = 0xFF;
				trafficAnnouncementCode = 0xFF;
				musicSpeechSwitchScode = 0xFF;
	  
				group0CodeBitsSteadyCount = 0;
				for(i = 0; i < sizeof(programmeServiceName); i++)
					programmeServiceName[i] = 0x20;
			  
			 
			  
				decoderIdentificationControlCode = 0xFF;
			
				for(i = 0; i < sizeof(alternativeFrequencyCodes); i++)
					alternativeFrequencyCodes[i] = 0;
			  
				linkageActuator = 0xFF;
			
				extendedCountryCode = 0;
				
				radioTextPosition = 0;
				radioTextAvailable = 0;
				radioFrequencyAvailable = 0;
			
				for(i = 0; i < sizeof(radioText); i++) 
				{
					radioText[i] = 0x20;
					radioTextPrevious[i] = 0x20;
				}
				
				for(i = 0; i < sizeof(radioTime); i++) 
				{
					radioTime[i] = 0x20;
				}
				
				for(i = 0; i < sizeof(radioFrequency); i++) 
				{
					radioFrequency[i] = 0x20;
				}
			
				textSegmentAddress0Seen = 0;
			  
				textVersionPrevious = 0;
				textSegmentAddressPrevious = 0;
			
				utcMinutesPrevious = 0xFF;
		  
			}

	  
		  //Programme Type code
			if (programmeTypeCode != ((block2 / 0x20) & 0x1F)) 
			{
				programmeTypeCode = ((block2 / 0x20) & 0x1F); //bits 5-9 (2^5)
				switch(programmeTypeCode) {
				case 0:
					PORTD &= ~_BV(PD0);// Nichts
					break;
				case 1:
					PORTD &= ~_BV(PD0);// Nachrichten
					break;
				case 2:
					PORTD &= ~_BV(PD0);// Aktuelles
					break;
				case 3:
					PORTD &= ~_BV(PD0);// Information
					break;
				case 4:
					PORTD &= ~_BV(PD0);// Sport
					break;
				case 5:
					PORTD &= ~_BV(PD0);// Erziehung
					break;
				case 6:
					PORTD &= ~_BV(PD0);// Drama
					break;
				case 7:
					PORTD &= ~_BV(PD0);// Kultur
					break;
				case 8:
					PORTD &= ~_BV(PD0);// Wissenschaft
					break;
				case 9:
					PORTD &= ~_BV(PD0);// Divers Sprache
					break;
				case 10:
					PORTD |= _BV(PD0);// Pop Musik
					break;
				case 11:
					PORTD |= _BV(PD0);// Rock Musik
					break;
				case 12:
					PORTD |= _BV(PD0);// Easy Listening
					break;
				case 13:
					PORTD |= _BV(PD0);// Leichte Klassik
					break;
				case 14:
					PORTD |= _BV(PD0);// Ernste Klassik
					break;
				case 15:
					PORTD &= ~_BV(PD0);// Andere Musik
					break;
				case 16:
					PORTD &= ~_BV(PD0);// Wetter
					break;
				case 17:
					PORTD &= ~_BV(PD0);// Finanzen
					break;
				case 18:
					PORTD &= ~_BV(PD0);// Kinder
					break;
				case 19: 
					PORTD &= ~_BV(PD0);// Soziales
					break;
				case 20:
					PORTD &= ~_BV(PD0);// Religion
					break;
				case 21:
					PORTD &= ~_BV(PD0);// Telefon
					break;
				case 22:
					PORTD &= ~_BV(PD0);// 3:Reise
					break;
				case 23:
					PORTD &= ~_BV(PD0);// Freizeit
					break;
				case 24:
					PORTD &= ~_BV(PD0);// Jazz Musik
					break;
				case 25:
					PORTD &= ~_BV(PD0);// Country Musik
					break;
				case 26:
					PORTD &= ~_BV(PD0);// Deutsche Musik
					break;
				case 27:
					PORTD |= _BV(PD0);// Oldies
					break;
				case 28:
					PORTD |= _BV(PD0);// Volksmusik
					break;
				case 29:
					PORTD |= _BV(PD0);// Dokumentation
					break;
				case 30:
					PORTD |= _BV(PD0);// Alarm Test
					break;
				case 31:
					PORTD |= _BV(PD0);// Alarm - Alarm !
					break;
				default:
					PORTD |= _BV(PD0);// Unbekannt
					break;
				}
			}
	  
	//Type 0 groups: Basic tuning and switching information
			if (groupType == 0) 
			{
	//Traffic announcement code > auf LED 5 gelb
				if (trafficAnnouncementCode != ((block2 / 0x10) & 0x01) ||
					trafficProgrammeIdentificationCode != (block2 & _BV(10)) / 0x400) 
				{
					trafficAnnouncementCode = (block2 / 0x10) & 0x01; //bit 4 (2^4)
					trafficProgrammeIdentificationCode = (block2 & _BV(10)) / 0x400; //bit 10
					if (trafficProgrammeIdentificationCode == 0) 
						 // LED5 gelb      
						PORTD |= _BV(PD1);
					else 
						PORTD &= ~_BV(PD1);  // LED5 gelb
				}
			  
	//Music Speech switch code > auf LED 6 gelb
				if (musicSpeechSwitchScode != ((block2 / 0x08) & 0x01)) 
				{
					musicSpeechSwitchScode = (block2 / 0x08) & 0x01;  //bit 3 (2^3)
					if (musicSpeechSwitchScode) 
						PORTD |= _BV(PD5);
					else
						PORTD &= ~_BV(PD5);
				}
			
	//Decode program service name and decoder identification control code
				group0CodeBits = block2 & 0x03; //0, 1, 2, 3;
			
			
			//TODO: improve decoderIdentificationControlCode detection, decouple from PS name
			//Decoder-identification control code-bit is bit 3 in block2
				if (group0CodeBits == 0) 
				{
					if (block2 & 0x04)
						decoderIdentificationControlCodeNew |= _BV(3);
					else 
						decoderIdentificationControlCodeNew &= ~_BV(3);
				}
				if (group0CodeBits == 1) 
				{
					if (block2 & 0x04)
						decoderIdentificationControlCodeNew |= _BV(2);
					else 
						decoderIdentificationControlCodeNew &= ~_BV(2);
				}
				if (group0CodeBits == 2) 
				{
					if (block2 & 0x04)
						decoderIdentificationControlCodeNew |= _BV(1);
					else 
						decoderIdentificationControlCodeNew &= ~_BV(1);
				}
				if (group0CodeBits == 3) 
				{
					if (block2 & 0x04)
						decoderIdentificationControlCodeNew |= _BV(0);
					else 
						decoderIdentificationControlCodeNew &= ~_BV(0);
				}
			 //fill in information
				i = group0CodeBits * 2;
				programmeServiceNameNew[i] = block4 / 0xFF; //bits 8-16 (2^8)
				programmeServiceNameNew[i + 1] = block4; //bit 0-8
				
				if (programmeServiceNameNew[i] != programmeServiceName[i] ||
					programmeServiceNameNew[i + 1] != programmeServiceName[i + 1]) 
				{
				  //detected change, reset counter if not already counting
					if (group0CodeBitsSteadyCount > 4)
						group0CodeBitsSteadyCount = 0;
				}
				
				//increase counter only when there are less then 4 received words
				if (group0CodeBitsSteadyCount < 4)
					group0CodeBitsSteadyCount++;
				
				programmeServiceName[i] = programmeServiceNameNew[i];
				programmeServiceName[i + 1] = programmeServiceNameNew[i + 1];
				
				//when we detected 4 new PS words then display it
//				if (group0CodeBitsSteadyCount == 4) 
				{
					lcd_setcursor(0, 1);
					for(i = 0; i < LCD_COLUMNS; i++)
					{
						if (i < 4)
							lcd_data(0x20);  // Leerzeichen
						else if (i < 12)	// 8 Zeichen zentriert
							lcd_data(programmeServiceName[i-4]);
						else
							lcd_data(0x20);  // Leerzeichen
					}
					group0CodeBitsSteadyCount++;
				  
					if (decoderIdentificationControlCode != decoderIdentificationControlCodeNew) 
					{
						decoderIdentificationControlCode = decoderIdentificationControlCodeNew;
						if (decoderIdentificationControlCode & 0b0001)
							PORTC |= _BV(PC1);  	// LED 3 grn
						else
							PORTC &= ~_BV(PC1);		// LED3 grn
					}
				}
			
	  
				if (groupVersion == 'A') 
				{
	//Alternative frequency codes
					for (h = 0; h < 2; h++) 
					{
						if (h == 0)
							j = block3; //first AF is in bits 0-7 of block3
						else
							j = block3 / 256; //second bits 8-15
					  
					//only frequencies we want, no control codes
						if (j >= 1 && j <= 204) 
						{	
							for(i = 0; i < sizeof(alternativeFrequencyCodes); i++) 
							{
								if (alternativeFrequencyCodes[i] == j) 
								{
									break;
								} 
								if (alternativeFrequencyCodes[i] == 0) 
								{
									alternativeFrequencyCodes[i] = j;
									break;
								}
							}
							j = alternativeFrequencyCodes[0];
							m = 875 + j;
							/*
							sprintf(radioFrequency,"%u.%u MHz", m / 10, m % 10);
							radioFrequencyAvailable = 1;		
							lcd_setcursor(0,2);
							for(i = 0; i < sizeof(radioFrequency); i++)
							{
								if (i==LCD_COLUMNS) 
									break;
								if (radioFrequency[i] = 0x00)
									radioFrequency[i] = 0x20;
								lcd_data(radioFrequency[i]);
							}
							*/
							
							
						}
						if (j == 224 && alternativeFrequencyCodes[25] != j) 
						{
							alternativeFrequencyCodes[25] = 224;
						}
					}
				}
			//version B contains PI code in block3
	  
			} 
			else if (groupType == 1) 
			{
	 //Type 1 groups: Programme Item Number and slow labelling codes     
				if (groupVersion == 'A') 
				{
				//TODO: 5 bit Radio Paging Codes, bits 0-4 block2
				//TODO: Add textual description of LA
				//Linkage Actuator
					if (linkageActuator != (block3 & _BV(15))) 
						linkageActuator = (block3 & _BV(15));
				//store variant code, bits 13-15
					i = (block3 / 0x2000) & 0x07;
				//TODO: Paging
					if (i == 0) 
					{
					//TODO: Add textual description of ECC code
					//Extended Country Code, bits 0-7 block3
					if (extendedCountryCode != (block3 & 0xFF)) 
						extendedCountryCode = (block3 & 0xFF);
			  
					} 
					
				}
			}
			else if (groupType == 2) 
			{
	//Type 2 groups: RadioText
			//text version A or B, bit 5 block 2
				if (block2 & 0x10) 
					textVersion = 'B';
				else
					textVersion = 'A';
			
			//block2 bit 0-3
				textSegmentAddress = (block2 & 0x0F);
			
			//clean radioText when version changes
				if (textVersionPrevious != 0 &&
					textVersionPrevious != textVersion) 
				{
					for(i = 0; i < sizeof(radioText); i++) 
						radioText[i] = 0;
					textSegmentAddressPrevious = 0;
					textSegmentAddress0Seen = 0;
				}
			
			//detected new start of text segment, normally address 0x00
				if (textSegmentAddressPrevious > textSegmentAddress) 
				{
					if (groupVersion == 'A')
						h = 64;
					else
						h = 32;
			  
			  //detect new radioText
					j = 0;
					for (i = 0; i < h; i++) 
					{
						if (radioText[i] != 0 && radioText[i] != ' ') 
						{
							if (radioText[i] != radioTextPrevious[i]) 
							{
								j = 1;
								break;
							}
						}
					}
				  
					//only print when we have received address 0 once.
					if (textSegmentAddress0Seen == 0)
						j = 0;
				   
					if (j) 
					{
						for (i = 0; i < h; i++) 
						{           
							if (radioText[i] == 0) 
							break;
							radioTextPrevious[i] = radioText[i];
						}
						radioTextAvailable = 1;	
						radioTextPosition = 0;	
						lcd_setcursor(0,2);
						for(i = 0; i < sizeof(radioText); i++)
						{
							if (i==LCD_COLUMNS) 
								break;
							if (radioText[i] == 0)
								break;
							lcd_data(radioText[i]);
						}
					}
				}
				else if (radioTextAvailable == 1)
				{
				// Scrollen
					lcd_setcursor(0,2);
					for(i = 0; i < sizeof(radioText); i++)
					{
						if (i==LCD_COLUMNS) 
							break;
						if (i+radioTextPosition > sizeof(radioText))
						{
							if (radioText[i+radioTextPosition-sizeof(radioText)] == 0)
								break;
							lcd_data(radioText[i+radioTextPosition-sizeof(radioText)]);
						}
						else
						{		
							if (radioText[i+radioTextPosition] == 0)
								break;
							lcd_data(radioText[i+radioTextPosition]);
						}	
					}
				}
				
				//64 bit messages in block 3 & 4
				if (groupVersion == 'A') 
				{
				  
					radioText[(textSegmentAddress * 4)] = block3 / 256;
					radioText[(textSegmentAddress * 4) + 1] = block3;
					radioText[(textSegmentAddress * 4) + 2] = block4 / 256;
					radioText[(textSegmentAddress * 4) + 3] = block4;
				  
				  //fill bytes smaller then textSegmentAddress with spaces when they are '0'
					if (textSegmentAddress > 0) 
					{
						for (i = 0; i < (textSegmentAddress - 1) * 4; i++) 
						{
							if (radioText[i] == 0)
							radioText[i] = ' ';
						}
					}
				  
				}
				//TODO: 32 bit messages not tested because of lack station transmitting it.
				//32 bit messages in block 4
				else 
				{
					radioText[(textSegmentAddress * 2)] = block4 / 256;
					radioText[(textSegmentAddress * 2) + 1] = block4;
				  
				  //fill bytes smaller then textSegmentAddress with spaces when they are '0'
					if (textSegmentAddress > 0) 
					{
						for (i = 0; i < (textSegmentAddress - 1) * 2; i++) 
						{
						if (radioText[i] == 0)
							radioText[i] = ' ';
						}
					}
				}
				if (textSegmentAddress == 0)
					textSegmentAddress0Seen = 1;
				textVersionPrevious = textVersion;
				textSegmentAddressPrevious = textSegmentAddress;
			} 
			else if (groupType == 3 && groupVersion == 'A') 
			{
				//Type 3A groups: Application identification for Open data		
			} 
			else if (groupType == 3 && groupVersion == 'B') 
			{
				//Type 3B groups: Open Data Application
			} 
			else if (groupType == 4 && groupVersion == 'A') 
	//Type 4A groups : Clock-time and date      
			{
			//bits 0-5 are in block4 as bits 6-11
				utcMinutes = (block4  / 64) & 0x3F;
				if (utcMinutesPrevious != utcMinutes) 
				{
					utcMinutesPrevious = utcMinutes;
			
				//bits 0-14 are in block3 as bits 1-15
				//bits 15-16 are in block2 as bits 0-1
					modifiedJulianDay = (block3 / 2) + (block2 & 0x03) * 32768;
			  
				//bits 0-3 are in block4 as bits 12-15
				//bit 4 is in block3 as bit 1
					utcHours = (block4 / 4096) + (block3 & 0x01) * 16;
			  
				//local time offset are bits 0-4 in block 4
					localTimeOffset = block4 & 0x1F;
				//sign is in bit 5 of block4, 0=+ 1=-
					if (block4 & 0x20) 
						localSign = '-';
					else
						localSign = '+';
				  
					//multiply by 30 so that we have offset in minutes (offset is in multiples of .5 hours)
					localTimeOffset *= 30;
		  
		//          printf_P(PSTR("CT: 0x%01X%04X%04X, "), (block2 & 0x03), block3, block4);
				  
					//Modified Julian date to year-month-day conversion
					utcYear = floor((modifiedJulianDay - 15078.2) / 365.25);
					utcMonth = floor((modifiedJulianDay - 14956.1 - floor(utcYear * 365.25)) / 30.6001);
					utcDay = modifiedJulianDay - 14956 - floor(utcYear * 365.25) - floor(utcMonth * 30.6001);
				
					if (utcMonth == 14 || utcMonth == 15)
						i = 1; 
					else 
						i = 0;
				
					utcYear = utcYear + i + 1900;
					utcMonth = utcMonth - 1 - (i * 12);
		  
		 //          sprintf(RDSData4, "4:%04u-%02u-%02u", utcYear, utcMonth, utcDay);
						   
				  //TODO: half hour timezones and negative timezones not tested because lack of station transmitting it.
				  //lets calulate local time
					if (localSign == '-') 
					{
						localHours = utcHours - (localTimeOffset / 60);
						localMinutes = utcMinutes - (localTimeOffset % 60);
					} 
					else 
					{
						localHours = utcHours + (localTimeOffset / 60);
						localMinutes = utcMinutes + (localTimeOffset % 60);
					}
				  
					if (localMinutes < 0) 
					{
						localMinutes += 60;
						localHours--;
					}
					
					if (localMinutes > 59) 
					{
						localMinutes -= 60;
						localHours++;
					}
					if (localHours < 0)
						localHours += 24;
					if (localHours > 23)
						localHours -= 24;
					if (radioTextAvailable == 0 && radioFrequencyAvailable == 0) 	
					  
					{
						lcd_setcursor(0,2);	
						sprintf(radioTime, "%02u:%02u:00\r\n", localHours, localMinutes);
						for(i = 0; i < sizeof(radioTime); i++)
						{
							if (i==LCD_COLUMNS) 
								break;
							if (radioTime[i] = 0x00)
								radioTime[i] = 0x20;	
							lcd_data(radioTime[i]);
						}
					}
				}
			} 
			else
		 
			if (groupType == 4 && groupVersion == 'B') 
			{
				//Type 4B groups: Open data application
			} else if (groupType == 5) 
			{
				//Type 5 groups: Transparent data channels or ODA
			} else if (groupType == 6) 
			{
				//Type 6 groups: In-house applications or ODA
			} else if (groupType == 7 && groupVersion == 'A') 
			{
				//Type 7A groups: Radio Paging or ODA
			} else if (groupType == 7 && groupVersion == 'B') 
			{
				//Type 7B groups: Open data application
			} else if (groupType == 8) 
			{
				//Type 8 groups: Traffic Message Channel or ODA
			} else if (groupType == 9) 
			{
				//Type 9 groups: Emergency warning systems or ODA
			} else if (groupType == 10 && groupVersion == 'A') 
			{
				//Type 10A groups: Programme Type Name
			} else if (groupType == 10 && groupVersion == 'A') 
			{
				//Type 10B groups: Open data
			} else if (groupType == 11) 
			{
				//Type 11 groups: Open Data Application
			} else if (groupType == 12) 
			{
				//Type 12 groups: Open Data Application
			} else if (groupType == 13 && groupVersion == 'A') 
			{
				//Type 13A groups: Enhanced Radio Paging or ODA
			} else if (groupType == 13 && groupVersion == 'B') 
			{
				//Type 13B groups: Open Data Application
			} else if (groupType == 14) 
			{
				//Type 14 groups: Enhanced Other Networks information
			} else if (groupType == 15 && groupVersion == 'A') 
			{
				//Type 15A groups: 'currently unavailable'
			} else if (groupType == 15 && groupVersion == 'B') 
			{
				//Type 15B groups: Fast basic tuning and switching information
			}
		} 
		else 
		{
		  sei();
		}
		
	}
}



























